%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Dynamic Syntax
%
% context-handling library
%
% Matthew Purver, 2004
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


:- dynamic context/1.

% clear context
reset_context :-
	retractall( context( _ ) ),
	assert( context( [] ) ).

% no context? do nothing
update_context( _Forest ) :-
	flag( context, none ).
% single turn? replace with complete forest
update_context( Forest ) :-
	flag( context, one ),
	retractall( context( _ ) ),
	complete_forest( Forest, Context ),
	assert( context( Context ) ).
% all turns? append
update_context( Forest ) :-
	flag( context, all ),
	context( Context1 ),
	retractall( context( _ ) ),
	complete_forest( Forest, Context ),
	append( Context, Context1, Context2 ),
	assert( context( Context2 ) ).

% any member of the context forest
tree_in_context( Tree ) :-
	context( Context ),
	member( twa( Tree, _Words, _Actions ), Context ).

% any node in a tree in context
node_in_context( Node ) :-
	tree_in_context( tree( Nodes, _Pointer ) ),
	member( Node, Nodes ).


% pronouns considered as always contextually available for generation
word_in_context( Word, Actions ) :-
	pron( Word, _, Type, _ ),
	Type \= whrel,
	Type \= whq,
	word( Word, Actions ).
% elliptical verbs always available
word_in_context( Word, Actions ) :-
	verb( _, Word, _, _, ellip ),
	word( Word, Actions ).
% otherwise any word/action pair in context
word_in_context( Word, [Action] ) :-
	context( Context ),
	member( twa( _Tree, Words, Actions ), Context ),
	nth( N, Words, Word ),
	nth( N, Actions, Action ).

% any sequence of word/action pairs in context
words_in_context( SubWords, SubActions ) :-
	context( Context ),
	member( twa( _Tree, Words, Actions ), Context ),
	sub_list( N, L, Words, SubWords ),
	sub_list( N, L, Actions, SubActions ).


% gets a sublist starting at N with length L>0
sub_list( N, L, List, SubList ) :-
	append( Prefix, Mid, List ),
	length( Prefix, P ),
	N is P + 1,
	append( SubList, _Suffix, Mid ),
	length( SubList, L ),
	L > 0.


% initial parse forest from context
forest_in_context( Forest ) :-
	%partial_forest( Forest1 ), % any contextual partial trees for shared utterances
	wh_forest( Forest ). % partial trees from wh-questions in context
		  

% initial forest for answer re-use give a wh-question tree in context
wh_forest( [twa( Adjusted, [], [] )] ) :-
	tree_in_context( Tree ),
	complete( Tree ),
	wh_adjust( Tree, Adjusted ),
	!.
wh_forest( [] ).

% % any tree with a wh-metavariable as an argument of the root proposition
% % -- actually, this ignores embedded wh-questions, which we want to be able to answer
% wh_question( tree( Nodes, Pointer ) ) :-
% 	root( Node ),
% 	member( node( Node, Labels ), Nodes ),
% 	member( fo(Formula), Labels ),
% 	Formula =.. [_Pred | Args],
% 	member( wh(Wh), Args ),
% 	var( Wh ).

% move pointer to existing wh-node, remove wh-including formulae
wh_adjust( tree( Nodes, _Pointer ), tree( Nodes2, Node ) ) :-
	member( node( Node, Labels ), Nodes ),
	member( fo(wh(Wh)), Labels ),
	remove_wh( wh(Wh), Nodes, Nodes2 ).

% remove wh-including formulae, and change type to requirement
remove_wh( _Wh, [], [] ).
remove_wh( Wh, [node( Node, Labels1 ) | T1], [node( Node, [?ty(Type) | Labels2] ) | T2] ) :-
	select( fo(Formula), Labels1, Labels ),
	exact_sub_term( Wh, Formula ),
	!,
	select( ty(Type), Labels, Labels2 ),
	remove_wh( Wh, T1, T2 ).
remove_wh( Wh, [Node | T1], [Node | T2] ) :-
	remove_wh( Wh, T1, T2 ).

exact_sub_term( X1, X2 ) :-
	X1 == X2.
exact_sub_term( X1, Formula ) :-
	X1 \= Formula,
	Formula =.. [_Pred | Args],
	member( X2, Args ),
	exact_sub_term( X1, X2 ).
